#include <errno.h>
#include <sys/types.h>
#include <sys/acct.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <grp.h>
#include <pwd.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>

#include "array.h"
#include "error.h"

int sig_abort = 0;

/* included in the source as this definition doesn't seem to be exported anywhere */
struct acct_v3
{
        char            ac_flag;                /* Flags */
        char            ac_version;             /* Always set to ACCT_VERSION */
        u_int16_t       ac_tty;                 /* Control Terminal */
        u_int32_t       ac_exitcode;            /* Exitcode */
        u_int32_t       ac_uid;                 /* Real User ID */
        u_int32_t       ac_gid;                 /* Real Group ID */
        u_int32_t       ac_pid;                 /* Process ID */
        u_int32_t       ac_ppid;                /* Parent Process ID */
        u_int32_t       ac_btime;               /* Process Creation Time */
        float           ac_etime;               /* Elapsed Time */
        comp_t          ac_utime;               /* User Time */
        comp_t          ac_stime;               /* System Time */
        comp_t          ac_mem;                 /* Average Memory Usage */
        comp_t          ac_io;                  /* Chars Transferred */
        comp_t          ac_rw;                  /* Blocks Read or Written */
        comp_t          ac_minflt;              /* Minor Pagefaults */
        comp_t          ac_majflt;              /* Major Pagefaults */
        comp_t          ac_swaps;               /* Number of Swaps */
        char            ac_comm[ACCT_COMM];     /* Command Name */
};

char *gen_temp_fname(int f_nr)
{
	char *buffer = (char *)malloc(128);
	if (!buffer)
		error_exit("out of memory - cannot allocate 128 bytes");

	sprintf(buffer, ".%d.%d._acctail", getpid(), f_nr);

	return buffer;
}

void sigh(int sig)
{
	sig_abort = 1;
}

unsigned long int expand_compt(comp_t val)
{
	unsigned long int exponent;
	unsigned long int mantisa;

	exponent = (val >> 13) & 07;
	mantisa = val & 017777;

	while (exponent-- > 0)
		mantisa <<= 3;

	return mantisa;
}

double compt_to_msec(comp_t timeval)
{
	return ((double)(expand_compt(timeval) * 1000)) / AHZ;
}

int start_file(char *new_file, char *old_file, int old_fd)
{
	int fd = open(new_file, O_RDWR | O_CREAT);
	if (fd == -1)
		error_exit("cannot create file %s", new_file);

	acct(NULL);
	if (acct(new_file) == -1)
		error_exit("cannot start accounting on file %s", new_file);

	if (old_file)
	{
		close(old_fd);
		if (unlink(old_file) == -1)
			error_exit("error deleting file %s", old_file);
	}

	return fd;
}

void show_v1_2(int delay, int statistics, int details)
{
	time_t start = 0, now;
	int f_nr = 0;
	int fd;
	char *fname = NULL;
	array	arr(9, 0);
	int loop;

	printf("COMMAND          FLGS UID     .GID      EXIT    UTIME   STIME   ETIME   MEM\n");
	if (details)
		printf("                 IO     RW     MINFLT MAJFLT SWAPS\n");
	printf("-------------------------------------------------------------------------------\n");

	for(;;)
	{
		struct acct record;
		int rc;
		char flags[5], nflags=0;
		struct passwd *pwent;
		struct group *grpent;
		char dummy[ACCT_COMM + 1];

		time(&now);

		if ((now - start) > delay || fd == -1)
		{
			char *dummy = gen_temp_fname(f_nr);

			f_nr = 1 - f_nr;

			fd = start_file(dummy, fname, fd);
			free(fname);
			fname = dummy;

			start = now;
		}

		rc = read(fd, &record, sizeof(record));

		if (sig_abort)
			break;

                if (rc == -1)
                {
                        if (errno != EINTR)
				error_exit("error reading");
                }
                else if (rc == 0)
                {
                        usleep(100000);
                        continue;
                }
                else if (rc != sizeof(record))
                        error_exit("short read while retrieving record (got: %d, expected: %d)", rc, sizeof(record));

		/* linux v2.6.x changed from ac_comm[ACCT_COMM+1] to
		 * ac_comm[ACCT_COMM]
		 */
		memcpy(dummy, record.ac_comm, ACCT_COMM);
		dummy[ACCT_COMM] = 0x00;

		if (record.ac_flag & AFORK)
			flags[nflags++] = 'F';
		if (record.ac_flag & ASU)
			flags[nflags++] = 'U';
		if (record.ac_flag & ACORE)
			flags[nflags++] = 'C';
		if (record.ac_flag & AXSIG)
			flags[nflags++] = 'S';
		flags[nflags] = 0x00;


		pwent = getpwuid(record.ac_uid);
		grpent = getgrgid(record.ac_gid);

		int cur_index = arr.addstring(dummy);
		arr.addcounter(cur_index, 0, compt_to_msec(record.ac_utime));
		arr.addcounter(cur_index, 1, compt_to_msec(record.ac_stime));
		arr.addcounter(cur_index, 2, compt_to_msec(record.ac_etime));
		arr.addcounter(cur_index, 3, expand_compt(record.ac_mem));
		arr.addcounter(cur_index, 4, expand_compt(record.ac_io));
		arr.addcounter(cur_index, 5, expand_compt(record.ac_rw));
		arr.addcounter(cur_index, 6, expand_compt(record.ac_minflt));
		arr.addcounter(cur_index, 7, expand_compt(record.ac_majflt));
		arr.addcounter(cur_index, 8, expand_compt(record.ac_swaps));

		printf("%-16s %-4s %-8s.%-8s %-5d %7.2f %7.2f %7.2f %6ld\n",
				dummy,
				flags,
				pwent -> pw_name,
				grpent -> gr_name,
				record.ac_exitcode,
				compt_to_msec(record.ac_utime),
				compt_to_msec(record.ac_stime),
				compt_to_msec(record.ac_etime),
				expand_compt(record.ac_mem)
		      );
		if (details)
		{
			printf("                 %-6ld %-6ld %-6ld %-6ld %-6ld\n",
					expand_compt(record.ac_io),
					expand_compt(record.ac_rw),
					expand_compt(record.ac_minflt),
					expand_compt(record.ac_majflt),
					expand_compt(record.ac_swaps)
			      );
		}
	}

	acct(NULL);
	if (fd != -1)
		close(fd);

	/* emit totals */
	printf("\n\n --- Totals ---\n");
	printf("COMMAND          UTIME      STIME      ETIME      MEM     IO      RW\n                 MINFLT  MAJFLT  SWAPS\n");
	printf("-------------------------------------------------------------------------------\n");

	for(loop=0; loop<arr.getN(); loop++)
	{
		printf("%-16s %7.2f %7.2f %7.2f %7f %7f %7f\n                 %7f %7f %7f\n", arr.getstring(loop),
				arr.getcounter(loop, 0),
				arr.getcounter(loop, 1),
				arr.getcounter(loop, 2),
				arr.getcounter(loop, 3),
				arr.getcounter(loop, 4),
				arr.getcounter(loop, 5),
				arr.getcounter(loop, 6),
				arr.getcounter(loop, 7),
				arr.getcounter(loop, 8)
		      );
	}

	if (fname)
	{
		if (unlink(fname) == -1)
			error_exit("failed to delete file %s", fname);
	}
}

void show_v3(int delay, int statistics, int details)
{
	time_t start = 0, now;
	int f_nr = 0;
	int fd;
	char *fname = NULL;
	array	arr(9, 0);
	int loop;

	printf("COMMAND          FLGS UID     .GID      EXIT    UTIME   STIME   ETIME   MEM\n");
	if (details)
	{
		printf("                 IO     RW     MINFLT MAJFLT SWAPS PID   PPID\n");
		printf("                 PROC_CREATED\n");
	}
	printf("-------------------------------------------------------------------------------\n");

	for(;;)
	{
		struct acct_v3 record;
		int rc;
		char flags[5], nflags=0;
		struct passwd *pwent;
		struct group *grpent;
		char dummy[ACCT_COMM + 1];

		time(&now);

		if ((now - start) > delay || fd == -1)
		{
			char *dummy = gen_temp_fname(f_nr);

			f_nr = 1 - f_nr;

			fd = start_file(dummy, fname, fd);
			free(fname);
			fname = dummy;

			start = now;
		}

		rc = read(fd, &record, sizeof(record));

		if (sig_abort)
			break;

                if (rc == -1)
                {
                        if (errno != EINTR)
				error_exit("error reading");
                }
                else if (rc == 0)
                {
                        usleep(100000);
                        continue;
                }
                else if (rc != sizeof(record))
                        error_exit("short read while retrieving record (got: %d, expected: %d)", rc, sizeof(record));

		/* linux v2.6.x changed from ac_comm[ACCT_COMM+1] to
		 * ac_comm[ACCT_COMM]
		 */
		memcpy(dummy, record.ac_comm, ACCT_COMM);
		dummy[ACCT_COMM] = 0x00;

		if (record.ac_flag & AFORK)
			flags[nflags++] = 'F';
		if (record.ac_flag & ASU)
			flags[nflags++] = 'U';
		if (record.ac_flag & ACORE)
			flags[nflags++] = 'C';
		if (record.ac_flag & AXSIG)
			flags[nflags++] = 'S';
		flags[nflags] = 0x00;


		pwent = getpwuid(record.ac_uid);
		grpent = getgrgid(record.ac_gid);

		int cur_index = arr.addstring(dummy);
		arr.addcounter(cur_index, 0, compt_to_msec(record.ac_utime));
		arr.addcounter(cur_index, 1, compt_to_msec(record.ac_stime));
		arr.addcounter(cur_index, 2, compt_to_msec(record.ac_etime));
		arr.addcounter(cur_index, 3, expand_compt(record.ac_mem));
		arr.addcounter(cur_index, 4, expand_compt(record.ac_io));
		arr.addcounter(cur_index, 5, expand_compt(record.ac_rw));
		arr.addcounter(cur_index, 6, expand_compt(record.ac_minflt));
		arr.addcounter(cur_index, 7, expand_compt(record.ac_majflt));
		arr.addcounter(cur_index, 8, expand_compt(record.ac_swaps));

		printf("%-16s %-4s %-8s.%-8s %-5d %7.2f %7.2f %7.2f %6ld\n",
				dummy,
				flags,
				pwent -> pw_name,
				grpent -> gr_name,
				record.ac_exitcode,
				compt_to_msec(record.ac_utime),
				compt_to_msec(record.ac_stime),
				compt_to_msec(record.ac_etime),
				expand_compt(record.ac_mem)
		      );
		if (details)
		{
			char *timestr = ctime((time_t *)&record.ac_btime);
			char *dummy = strchr(timestr, '\n');
			if (dummy) *dummy = 0x00;	

			printf("                 %-6ld %-6ld %-6ld %-6ld %-6ld %-5ld %-5ld\n",
					expand_compt(record.ac_io),
					expand_compt(record.ac_rw),
					expand_compt(record.ac_minflt),
					expand_compt(record.ac_majflt),
					expand_compt(record.ac_swaps),
					record.ac_pid,
					record.ac_ppid
			      );
			printf("                 %-62s\n", timestr);
		}
	}

	acct(NULL);
	if (fd != -1)
		close(fd);

	/* emit totals */
	printf("\n\n --- Totals ---\n");
	printf("COMMAND          UTIME      STIME      ETIME      MEM     IO      RW\n                 MINFLT  MAJFLT  SWAPS\n");
	printf("-------------------------------------------------------------------------------\n");

	for(loop=0; loop<arr.getN(); loop++)
	{
		printf("%-16s %7.2f %7.2f %7.2f %7f %7f %7f\n                 %7f %7f %7f\n", arr.getstring(loop),
				arr.getcounter(loop, 0),
				arr.getcounter(loop, 1),
				arr.getcounter(loop, 2),
				arr.getcounter(loop, 3),
				arr.getcounter(loop, 4),
				arr.getcounter(loop, 5),
				arr.getcounter(loop, 6),
				arr.getcounter(loop, 7),
				arr.getcounter(loop, 8)
		      );
	}

	if (fname)
	{
		if (unlink(fname) == -1)
			error_exit("failed to delete file %s", fname);
	}
}

void probe_version(char *me)
{
	char *fname = gen_temp_fname(0);
	int fd = start_file(fname, NULL, -1);
	char buffer[2];
	pid_t child = fork();
	int rc;

	if (child == -1)
		error_exit("error forking");

	if (child == 0)
	{
		if (execlp(me, me, "-X", (void *)NULL))
			error_exit("failed to invoke myself");

		exit(1);
	}

	for(;;)
	{
		rc = read(fd, buffer, 2);
		if (rc != 0)
			break;

		sleep(1);
	}
	if (rc != 2)
		error_exit("Short-read reading from bufferfile");

	printf("Version: %d\n", buffer[1]);

	acct(NULL);
}

void usage(void)
{
	fprintf(stderr, "-s  time before the temporary file is deleted\n");
	fprintf(stderr, "-d  show details\n");
	fprintf(stderr, "-3  input is v3\n");
	fprintf(stderr, "-p  probe input version\n");
	fprintf(stderr, "-h  this help\n");
}

void version(void)
{
	fprintf(stderr, "acctail v" VERSION ", (C) folkert@vanheusden.com\n\n");
}

int main(int argc, char *argv[])
{
	int delay = 3600;
	int c;
	int statistics = 0;
	int details = 0;
	int v_3 = 0;

	while((c = getopt(argc, argv, "VXs:3dhtvp")) != -1)
	{
		switch(c)
		{
			case 'X':
				return 127;

			case '3':
				v_3 = 1;
				break;

			case 's':
				delay = atoi(optarg);
				break;

			case 't':
				statistics = 1;
				break;

			case 'v':
				statistics = 2;
				break;

			case 'p':
				probe_version(argv[0]);
				return 0;

			case 'd':
				details = 1;
				break;
			
			case 'V':
				version();
				return 0;

			case 'h':
			default:
				usage();
				return 1;
		}
	}

	version();

	signal(SIGTERM, sigh);
	signal(SIGINT, sigh);

	if (v_3)
		show_v3(delay, statistics, details);
	else
		show_v1_2(delay, statistics, details);

	return 0;
}
